/*-------------------------------------------------------------------------+
| Copyright 2013 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstChildWithType;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.ecore.util.FeatureMap;
import org.eclipse.emf.ecore.xml.type.AnyType;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.IMigrationProvider;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.introspection.items.MigrationServiceIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IMigrationService;
import org.fortiss.tooling.kernel.service.base.ObjectAwareServiceBase;

/**
 * This class implements the {@link IMigrationService} interface.
 * 
 * @author mou
 */
public class MigrationService extends ObjectAwareServiceBase<IMigrationProvider>
		implements IMigrationService, IIntrospectiveKernelService {
	/** The singleton instance. */
	private static final MigrationService INSTANCE = new MigrationService();

	/** Returns singleton instance of the service. */
	public static MigrationService getInstance() {
		return INSTANCE;
	}

	/** The connector extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.migrationProvider";

	/** The connector configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "migrationProvider";

	/** The connector attribute name. */
	private static final String ATTRIBUTE_NAME = "migrationProvider";

	/** {@inheritDoc} */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** Registers the migration provider with the service. */
	@Override
	public void registerMigrationProvider(IMigrationProvider provider, Class<?> modelElementClass) {
		addHandler(modelElementClass, provider);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service allows the registration of migrator classes, which perform the" +
				"\nmigration of models should the meta-model of the modeling language change." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/** {@inheritDoc} */
	@Override
	public boolean needMigration(ITopLevelElement input, Map<EObject, AnyType> unknownFeatures) {
		for(IMigrationProvider provider : getProviders(input)) {
			if(provider.needMigration(input, unknownFeatures)) {
				return true;
			}
		}
		return false;
	}

	/** Extracts the list of feature names from a map of features. */
	private List<String> featuresToStrings(Map<EObject, AnyType> features) {
		List<String> res = new BasicEList<String>();
		for(AnyType featuresPerKey : features.values()) {
			FeatureMap actualFeatures = featuresPerKey.getMixed();
			actualFeatures.addAll(featuresPerKey.getAnyAttribute());
			for(FeatureMap.Entry feature : actualFeatures) {
				res.add(feature.getEStructuralFeature().getName());
			}
		}
		return res;
	}

	/** {@inheritDoc} */
	@Override
	public void migrate(final ITopLevelElement input, final Map<EObject, AnyType> unknownFeatures) {
		input.runAsNonDirtyingCommand(new Runnable() {

			@Override
			public void run() {
				Map<EObject, AnyType> featuresNotMigrated = new HashMap<EObject, AnyType>();
				for(IMigrationProvider provider : getProviders(input)) {
					if(provider.needMigration(input, unknownFeatures)) {
						featuresNotMigrated.putAll(provider.migrate(input, unknownFeatures));
					}
				}
				for(IMigrationProvider provider : getProviders(input)) {
					if(provider.needMigration(input, featuresNotMigrated)) {
						error(ToolingKernelActivator.getDefault(),
								"Migrating " + input.getSaveableName() + " failed: A need for " +
										"migration is indicated although the migration was " +
										"already performed. Please fix the migrator (or model).");
					}
				}
				if(!featuresNotMigrated.isEmpty()) {
					error(ToolingKernelActivator.getDefault(), input.getSaveableName() +
							" contains one or more unknown feature(s) which will be deleted: " +
							featuresToStrings(unknownFeatures));

					for(Entry<EObject, AnyType> e : featuresNotMigrated.entrySet()) {
						EcoreUtil.delete(e.getKey(), true);
					}
				}
			}
		});

		try {
			input.doSave(new NullProgressMonitor());
		} catch(Exception e) {
			error(ToolingKernelActivator.getDefault(), "Error during migration of model!", e);

			if(e instanceof RuntimeException) {
				throw(RuntimeException)e;
			}
			throw new RuntimeException(e);
		}
	}

	/** Get all suitable {@link IMigrationProvider} for the given input. */
	private List<IMigrationProvider> getProviders(ITopLevelElement input) {
		List<IMigrationProvider> providers = new ArrayList<>();
		EObject migElement = input.getRootModelElement();

		for(Entry<Class<?>, List<IMigrationProvider>> migEntry : handlerMap.entrySet()) {
			if(getFirstChildWithType(migElement, migEntry.getKey()) != null) {
				providers.addAll(migEntry.getValue());
			}
		}

		return providers;
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Migration Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return new MigrationServiceIntrospectionDetailsItem(handlerMap);
	}
}
